/*
Stallation version 0.7  (2000-06-28)

Written by Peter Folk <pfolk@uni.uiuc.edu>

The contents of this file are in the public domain, but
the author would appreciate it if any useful changes or
corrections were communicated back to him, for inclusion
in a future version.
*/
#include "Stallation.h"
#include <vector>
#include <map>
#include <set>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <be/StorageKit.h>
#include <be/kernel/fs_attr.h>
#include <be/kernel/fs_info.h>
#include <be/support/String.h>


map<uint64,BString> registry;
set< pair<BString,uint64> > rregistry;

static status_t LockRegistry();
static status_t LockRegistry()
{
	return B_OK;
}
static status_t UnlockRegistry();
static status_t UnlockRegistry()
{
	return B_OK;
}
static status_t LoadRegistry();
static status_t LoadRegistry()
{
	cc* name = "Stallations";
	registry.clear();
	rregistry.clear();
	
	BPath path;
	status_t s = find_directory(B_USER_SETTINGS_DIRECTORY, &path);
	if (s != B_OK) return s;
	
	path.Append(name);
	FILE *f = fopen(path.Path(), "r");
	if (f==NULL) return errno;
	
	// read the registry into the message
	while (!feof(f))
	{
		char c='\0';
		uint64 tag;
		
		if (fscanf(f,"%Lx",&tag)==1 && (c=getc(f))==' ') // skip space
		{
			BString name;
			while ((c=getc(f)) && c!=EOF && c!='\n')
				name += c;
			
			registry[tag]=name;
			rregistry.insert(make_pair(name,tag));
		}
		while (c!='\n' && !feof(f)) c=getc(f);
	}
	
	return B_OK;
}
static status_t SaveRegistry();
static status_t SaveRegistry()
{
	cc* name = "Stallations";
	BPath path;
	status_t status = find_directory(B_USER_SETTINGS_DIRECTORY, &path);
	if (status != B_OK) return status;
	
	
	path.Append(name);
	FILE *f = fopen(path.Path(), "w");
	if (f==NULL) return errno;
	
	// read the registry into the message
	for (map<uint64,BString>::iterator i=registry.begin();
		i!=registry.end(); ++i)
	fprintf(f, "%Lx %s\n",i->first, i->second.String());
	
	return B_OK;
}



Stallation::Stallation(cc* name, bool reg)
: BStringItem(name), tag(real_time_clock_usecs())
{
	tag += uint64(getpid())<<48;
	// it's unlikely two installations will be created in the
	// same millisecond on one computer.  It's VERY unlikely
	// they will be created in the same millisecond on differ-
	// ent computers by processes with the same pid.
	
	sprintf(str,"%016Lx",tag);
	if (reg) Register();
}

Stallation::Stallation(cc* name, uint64 t, bool reg=false)
: BStringItem(name), tag(t)
{
	sprintf(str,"%016Lx",tag);
	if (reg) Register();
}

Stallation::~Stallation(){}


size_t Stallation::Count(cc* name)
{
	LoadRegistry();
	set< pair<BString,uint64> >::iterator i;
	i = rregistry.lower_bound(make_pair(BString(name),0));
	
	int j=0;
	while (i != rregistry.end() && i->first == name) ++j;
	return j;
}

Stallation* Stallation::Find(cc* name, int index)
{
	LoadRegistry();
	
	set< pair<BString,uint64> >::iterator i;
	if (name)
	{
		i = rregistry.lower_bound(make_pair(BString(name),0));
		while (--index>=0) ++i;
		if (i == rregistry.end() || i->first != name)
			return NULL;
	}
	else
	{
		i = rregistry.begin();
		while (--index>=0) ++i;
		if (i == rregistry.end())
			return NULL;
	}
	
	return new Stallation(i->first.String(),i->second);
}

Stallation* Stallation::Find(uint64 tag)
{
	map<uint64,BString>::iterator i = registry.find(tag);
	if (i == registry.end()) return NULL;
	else return new Stallation(i->second.String(),i->first);
}


void Stallation::Register()
{
	LockRegistry();
	LoadRegistry();
	registry[tag] = Name();
	rregistry.insert( make_pair(BString(Text()),tag) );
	SaveRegistry();
	UnlockRegistry();
}
void Stallation::SetName(cc* name)
{
	Unregister();
	SetText(name);
	Register();
}
void Stallation::Unregister()
{
	LockRegistry();
	LoadRegistry();
	
	map<uint64,BString>::iterator i = registry.find(tag);
	if (i != registry.end()) registry.erase(i);
	
	set< pair<BString,uint64> >::iterator j;
	j = rregistry.lower_bound( make_pair(BString(Name()),0) );
	while (j!=rregistry.end() && j->first==Name() && j->second!=tag) ++j;
	if (j != rregistry.end()) rregistry.erase(j);
	
	SaveRegistry();
	UnlockRegistry();
}


void Stallation::Install(const entry_ref &ref, bool interact)
{
	AddRef(ref,true);
//	if (ref/INSTALL exists)
//		INSTALL(ref,interact);
}

void Stallation::AddRef(const entry_ref &ref, bool recursive)
{
	BNode n(&ref);
	attr_info a;
	BString s;
	
	
	if (n.GetAttrInfo(STALL_ATTR,&a) == B_OK)
	{
		char* p = s.LockBuffer(a.size+1);
		a.size = n.ReadAttr(STALL_ATTR,B_STRING_TYPE,0,p,a.size);
		p[a.size]='\0';
		s.UnlockBuffer(strlen(p));
	}
	
	if (s.IFindFirst(String())==B_ERROR)
	{
		s.Append(",").Append(String());
		
		s.ReplaceAll(",,",",");
		while (s[0]==',') s.Remove(0,1);
		
		n.WriteAttr(STALL_ATTR,B_STRING_TYPE,0,s.String(),s.Length()+1);
	}
	
	if (recursive)
	{
		BDirectory d(&ref);
		entry_ref child;
		while (d.GetNextRef(&child)==B_OK)
			AddRef(child,recursive);
	}
}

bool Stallation::HasRef(const entry_ref &ref) const
{
	BNode n(&ref);
	attr_info a;
	
	if (n.GetAttrInfo(STALL_ATTR,&a) != B_OK) return false;
	
	
	BString s;
	char* p = s.LockBuffer(a.size+2);
	if (n.ReadAttr(STALL_ATTR,B_STRING_TYPE,0, p,a.size) > 0)
	{
		p[a.size]=',';
		p[a.size+1]='\0';
	}
	else p[0]='\0';
	
	s.UnlockBuffer(strlen(p));
	
	return s.IFindFirst(String()) != B_ERROR;
}


void Stallation::SubsumeRef(const entry_ref &ref, Stallation *inst)
{
	BNode n(&ref);
	if (n.InitCheck() != B_OK) return;
	
	attr_info a;
	BString s;
	
	
	if (n.GetAttrInfo(STALL_ATTR,&a) == B_OK)
	{
		char* p = s.LockBuffer(a.size+1);
		a.size = n.ReadAttr(STALL_ATTR,B_STRING_TYPE,0,p,a.size);
		p[a.size]='\0';
		s.UnlockBuffer(strlen(p));
	}
	
	// remove String() from s
	s.IReplaceAll(inst->String(),"");
	s.ReplaceAll(",,",",");
	
	if (s.IFindFirst(String())==B_ERROR)
	{
		s.Append(",").Append(String());
		s.ReplaceAll(",,",",");
		while (s[0]==',') s.Remove(0,1);
	}
	
	n.WriteAttr(STALL_ATTR,B_STRING_TYPE,0, s.String(),s.Length()+1);
}

void Stallation::SubsumeAll(dev_t dev, Stallation* inst)
{
	BVolume vol(dev);
	BQuery q;
	q.SetVolume(&vol);
	q.PushAttr(STALL_ATTR);
	q.PushString(inst->String());
	q.PushOp(B_CONTAINS);
	q.Fetch();
	
	entry_ref ref;
	while (q.GetNextRef(&ref) == B_OK)
		SubsumeRef(ref,inst);
}

void Stallation::SubsumeAll(Stallation* inst)
{
	dev_t dev;
	int32 i=0;
	while ((dev=next_dev(&i)) >= 0)
		SubsumeAll(dev,inst);
}


void Stallation::RemoveRef(const entry_ref &ref, bool recursive, bool nuke)
{
	BNode n(&ref);
	if (n.InitCheck() != B_OK) return;
	
	attr_info a;
	BString s;
	bool wasmine=false;
	
	if (n.GetAttrInfo(STALL_ATTR,&a) == B_OK)
	{
		char* p = s.LockBuffer(a.size+1);
		a.size = n.ReadAttr(STALL_ATTR,B_STRING_TYPE,0, p,a.size);
		p[a.size]='\0';
		s.UnlockBuffer(strlen(p));
		
		// remove String() from s
		ssize_t len = s.Length();
		s.IReplaceAll(String(),"");
		wasmine = (s.Length() != len);
		
		s.ReplaceAll(",,",",");
		while (s[s.Length()-1]==',') s.Remove(s.Length()-1,1);
		while (s[0]==',') s.Remove(0,1);
		
		if (s.Length() != len)
		{
			if (s.Length()==0) n.RemoveAttr(STALL_ATTR);
			else n.WriteAttr(STALL_ATTR,B_STRING_TYPE,0,
						s.String(),s.Length()+1);
		}
	}
	
	if (recursive)
	{
		BDirectory d(&ref);
		entry_ref child;
		while (d.GetNextRef(&child)==B_OK)
			RemoveRef(child,recursive,nuke);
	}
	
	if (nuke && wasmine && s.Length()==0)
		BEntry(&ref).Remove();
}

void Stallation::RemoveAll(dev_t dev, bool nuke)
{
	BVolume vol(dev);
	BQuery q;
	q.SetVolume(&vol);
	q.PushAttr(STALL_ATTR);
	q.PushString(String());
	q.PushOp(B_CONTAINS);
	q.Fetch();
	
	entry_ref ref;
	while (q.GetNextRef(&ref) == B_OK)
		RemoveRef(ref,true,nuke);
}

void Stallation::RemoveAll(bool free_if_unclaimed)
{
	dev_t dev;
	int32 i=0;
	while ((dev=next_dev(&i)) >= 0)
		RemoveAll(dev,free_if_unclaimed);
}
